#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from __future__ import annotations

from typing import Any

# Validation errors are a separate subset of TypeError to handle the
# custom type validation defined for complex object data types and allow
# runtime type confidence.


class ValidationError(TypeError): ...


class TypeValidationError(ValidationError):
    def __init__(self, instance: Any, expected: type[Any] | None):
        self.typ_inp: str = type(instance).__name__
        self.typ_out: str = expected.__name__ if isinstance(expected, type) else "None"

    def __str__(self) -> str:
        return "Instance has invalid type, provided type '%s' but expected '%s'." % (
            self.typ_inp,
            self.typ_out,
        )


class PendingAiError(Exception): ...


# -----


class HttpError(PendingAiError): ...


class RequestError(HttpError):
    def __str__(self) -> str:
        return "A request to the Pending AI server failed. Please try again shortly or contact support via email at 'support@pending.ai'."


class RequestTimeoutError(HttpError):
    def __str__(self) -> str:
        return "A request to the Pending AI server timed out. Please try again shortly."


# -----


class AuthenticationError(PendingAiError): ...


class ClientError(PendingAiError):
    def __init__(self, service: str | None = None, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.service: str | None = service

    def __str__(self) -> str:
        if self.service:
            return (
                "An error occurred while handling your request to "
                f"the Pending AI '{self.service.title()}' service."
            )
        return "An error occurred while handling your request to the Pending AI service."


class ClientTimeout(ClientError): ...


class AuthError(PendingAiError): ...


class UnexpectedResponseError(PendingAiError):
    def __init__(self, method: str, name: str, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.method: str = method
        self.name: str = name

    def __str__(self) -> str:
        return (
            f"An unexpected error occurred from a {self.method.upper()} "
            f"'{self.name}' request. Please try again and if the "
            "issue persists contact Pending AI support via email "
            "at 'support@pending.ai'."
        )


class NotFoundError(ClientError):
    def __init__(
        self,
        resource_id: str,
        resource_name: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.resource_id: str = resource_id
        self.resource_name: str = resource_name

    def __str__(self) -> str:
        return (
            f"The requested '{self.resource_name.title()}' "
            "resource cannot be retrieved as it does not exist: "
            f"'{self.resource_id}'."
        )


class PaymentGatewayError(ClientError):
    def __str__(self) -> str:
        return (
            "A billing service error was encountered while handling "
            "your request. No changes were made during the process, "
            "please try again. If the error persists, contact Pending "
            "AI support via email at 'support@pending.ai'."
        )


class ServiceUnavailableError(ClientError):
    def __str__(self) -> str:
        return (
            "The requested service is currently unavailable, please "
            "try again shortly. If the error persists, contact Pending "
            "AI support via email at 'support@pending.ai'."
        )


class ContentTooLargeError(ClientError):
    def __init__(self, msg: str, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.msg: str = msg

    def __str__(self) -> str:
        return self.msg


class RequestValidationError(ClientError):
    def __init__(self, errors: list[dict], *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.errors: list[dict] = errors

    def __str__(self) -> str:
        return f"Invalid parameters provided for service request: {self.errors}."


class UnauthorizedError(ClientError):
    """
    HTTP response has status `401 Unauthorized`.
    """

    def __str__(self) -> str:
        return (
            "Unauthorized access to Pending AI services. Ensure an "
            "authenticated api key is provided to the Pending AI "
            "client. For further help, contact support via email at "
            "'support@pending.ai'."
        )


class ForbiddenError(ClientError):
    """
    HTTP response has status `403 Forbidden`.
    """

    def __str__(self) -> str:
        return (
            "Forbidden access to Pending AI services. The provided "
            "api key does not have sufficient permissions to access "
            "the requested services. For further help, contact support "
            "via email at 'support@pending.ai'."
        )


class UnsubscribedError(ClientError):
    """
    HTTP response has status `403 Forbidden` and `not_subscribed` in the
    response content.
    """

    def __str__(self) -> str:
        return (
            "Forbidden access to the requested Pending AI service due "
            "to no active subscription. For further help, reach out "
            "to Pending AI support via email at 'support@pending.ai' "
            "and setup a new subscription."
        )


class HtmlGenerationError(PendingAiError):
    """This error is raised if the HTML cannot be generated."""

    def __str__(self) -> str:
        original_message: str = super().__str__()
        return f"Failed to generate HTML report: {original_message}"



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from __future__ import annotations

from pendingai.requestor.requestor import Requestor
from pendingai.requestor.response import RequestorResponse

__all__: list[str] = ["Requestor", "RequestorResponse"]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from __future__ import annotations

import os
from datetime import datetime, timedelta

from pendingai import config
from pendingai.abc import Singleton
from pendingai.auth.session_token import SessionToken
from pendingai.exceptions import AuthenticationError
from pendingai.utils.context import Context
from pendingai.utils.logger import Logger

logger = Logger().get_logger()


class SessionInfo:
    """
    Session info attached to a session token.
    """

    def __init__(self, token: SessionToken):
        self.issued_at: datetime = token.claims.issued_at
        self.expire_at: datetime = token.claims.expire_at
        dt: datetime = datetime.now().astimezone(config.TZ_LOCAL)
        self.total_time: timedelta = self.expire_at - self.issued_at
        self.remaining: timedelta = self.expire_at - dt
        self.user_email: str = token.claims.user_email
        self.user_org_name: str = token.claims.user_org_name


class AuthSession(Singleton):
    """
    Session authentication manager persisting session token handling.
    """

    info: SessionInfo | None = None
    token: SessionToken | None = None

    def _initialize(self, *, access_token: str | None = None) -> None:
        # extract cached session token information from runtime context
        context = Context()
        token_cache: SessionToken | None = context.cache.session_token

        # setup a session with input authentication from different input
        # sources with precedence cache hit, input argument, environment
        # variable, and empty otherwise
        if access_token and token_cache and token_cache.access_token == access_token:
            logger.debug("Creating session with token from token cache")
            self.update_session(token_cache)

        elif access_token:
            logger.debug("Creating session with new token from input")
            self.update_session(SessionToken(access_token=access_token))

        elif token_cache:
            logger.debug("Creating session with token from token cache")
            self.update_session(token_cache)

        elif access_token := os.environ.get("PENDINGAI_TOKEN"):
            logger.debug("Creating session with new token from environment")
            self.update_session(SessionToken(access_token=access_token))

        # for a session with an initialized session token, global cache
        # data must be updated and saved to overwrite existing sessions
        if self.token:
            if token_cache and token_cache.access_token != self.token.access_token:
                logger.info("New auth session token saving to cache")
                context.cache.session_token = self.token
                context.cache.save()

    def update_session(self, token: SessionToken):
        """
        Update session and session info with a new token.
        """
        if token.is_expired():
            raise AuthenticationError("Session has expired, please log in again.")
        self.token = token
        self.info = SessionInfo(self.token)

    def remove_session(self):
        """
        Remove session and session info.
        """
        self.token = None
        self.info = None



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import enum


@enum.unique
class Environment(str, enum.Enum):
    """
    Deployment environment used for building client connection strings,
    authentication flows for the device or refresh tokens, and controls
    cached state data between different environments
    """

    DEV = "dev"
    STAGING = "stage"
    DEFAULT = "default"



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from __future__ import annotations

from requests import Response

from pendingai.api_resources.interfaces import (
    ListResourceInterface,
    RetrieveResourceInterface,
)
from pendingai.api_resources.object import ListObject, Object
from pendingai.api_resources.parser import cast
from pendingai.exceptions import (
    NotFoundError,
    ServiceUnavailableError,
    UnexpectedResponseError,
)


class Model(Object):
    """
    Model object.
    """

    id: str
    """
    Resource id.
    """
    object: str = "model"
    """
    Resource object type.
    """
    name: str | None
    """
    Optional name of the model.
    """
    description: str | None
    """
    Optional description of the model.
    """
    version: str | None
    """
    Optional version of the model.
    """
    summary: dict
    """
    Optional summary statistics of the model.
    """
    metadata: dict
    """
    Optional metadata describing specific model features.
    """


class ModelStatus(Object):
    """
    Model status object.
    """

    status: str


class ModelInterface(
    ListResourceInterface[Model],
    RetrieveResourceInterface[Model],
):
    """
    Model resource interface; utility methods for model resources.
    """

    def list(
        self,
        *,
        limit: int = 25,
        next_page: str | None = None,
        prev_page: str | None = None,
    ) -> ListObject[Model]:
        if next_page and prev_page:
            raise ValueError("Cannot specify both next_page and prev_page")
        if limit < 1 or limit > 100:
            raise ValueError("List 'limit' must be in range: [1, 100].")
        params: dict[str, str | int] = {"limit": limit}
        if next_page is not None:
            params["next-page"] = next_page
        if prev_page is not None:
            params["prev-page"] = prev_page

        r: Response = self._requestor.request(
            "GET", "/generator/v1/models", params=params
        )
        if r.status_code == 200:
            return cast(ListObject[Model], r.json())
        raise UnexpectedResponseError("GET", "list_model")

    def retrieve(self, id: str, *args, **kwargs) -> Model:
        r: Response = self._requestor.request("GET", f"/generator/v1/models/{id}")
        if r.status_code == 200:
            return cast(Model, r.json())
        elif r.status_code == 404:
            raise NotFoundError(id, "Model")
        elif r.status_code == 503:
            raise ServiceUnavailableError
        raise UnexpectedResponseError("GET", "retrieve_model")

    def status(self, id: str) -> ModelStatus:
        r: Response = self._requestor.request("GET", f"/generator/v1/models/{id}/status")
        if r.status_code == 200:
            return cast(ModelStatus, r.json())
        elif r.status_code == 404:
            raise NotFoundError(id, "Model")
        elif r.status_code == 503:
            raise ServiceUnavailableError
        raise UnexpectedResponseError("GET", "retrieve_model_status")



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from typing import Any


class Singleton(object):
    """
    Abstract singleton metaclass.
    """

    _instance = None

    def __new__(cls, *args: Any, **kwargs: Any):
        if not cls._instance:
            if not cls._instance:
                cls._instance = super().__new__(cls)
                cls._instance._initialize(*args, **kwargs)
        return cls._instance

    def _initialize(self, *_: Any, **__: Any) -> None:
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from __future__ import annotations

from typer import Exit, Typer

from pendingai.auth import SessionInfo
from pendingai.auth.session import AuthSession
from pendingai.cli.console import Console
from pendingai.cli.context import PendingAiContext
from pendingai.cli.utils import catch_exception
from pendingai.utils.logger import Logger

logger = Logger().get_logger()
cout = Console()
app = Typer(
    name="auth",
    help="Manage an authenticated session with the Pending AI platform.",
    short_help="Authenticate with the Pending AI platform.",
    no_args_is_help=True,
)


# region command: login ------------------------------------------------


@app.command("login", help="Login to a Pending AI account.")
@catch_exception()
def _login(ctx: PendingAiContext):
    ctx.obj["client"].authentication.login()
    info: SessionInfo | None = ctx.obj["client"].session.info
    cout.print("[green]✓ Authentication complete.")
    if info:
        cout.print(f"[green]✓ Logged in user account: [bold]{info.user_email}")
        cout.print(f"[green]✓ Logged in organization: [bold]{info.user_org_name}")


# region command: logout -----------------------------------------------


@app.command("logout", help="Logout of a Pending AI account.")
@catch_exception()
def _logout(ctx: PendingAiContext):
    info: SessionInfo | None = ctx.obj["client"].session.info
    if info:
        ctx.obj["client"].authentication.logout()
        cout.print(f"[success]✓ Logged out [b underline]{info.user_email}[/].")
    else:
        cout.print("[warn]! Already logged out.")


# region command: refresh ----------------------------------------------


@app.command("refresh", help="Refresh an authenticated session.")
@catch_exception()
def _refresh(ctx: PendingAiContext):
    info: SessionInfo | None = ctx.obj["client"].session.info
    if not info:
        cout.print("[warn]! No logged in user, use [code]pendingai auth login[/].")
        raise Exit(1)
    ctx.obj["client"].authentication.refresh()
    cout.print("[green]✓ Authentication complete.")
    if info:
        cout.print(f"[green]✓ Logged in as user: [bold]{info.user_email}")
        cout.print(f"[green]✓ Logged in to team: [bold]{info.user_org_name}")


# region command: status -----------------------------------------------


@app.command("status", help="Show active session information.")
@catch_exception()
def _status(ctx: PendingAiContext):
    session: AuthSession = ctx.obj["client"].session
    if not session.info:
        cout.print("[warn]! No logged in user, use [code]pendingai auth login[/].")
        raise Exit(1)
    elif session.token and session.token.is_expired():
        cout.print("[warn]! Access has expired, use [code]pendingai auth refresh[/].")
        raise Exit(1)

    if session.info.remaining.days >= 1:
        cout.print("[success]✓ Session is active: Over 1 day remaining")
    elif session.info.remaining.total_seconds() < 60:
        cout.print("[warn]! Session is closing soon: Less than 1 minute remaining")
    else:
        time_left: str = str(session.info.remaining).split(".")[0]
        cout.print(f"[success]✓ Session is active: [not b]{time_left} remaining")
    cout.print(f"- Session user account: [b underline]{session.info.user_email}")
    cout.print(f"- Session organization: [b]{session.info.user_org_name}")


# region command: token ------------------------------------------------


@app.command("token", help="Get the access token of the current session.")
@catch_exception()
def _token(ctx: PendingAiContext):
    if not ctx.obj["client"].session.token:
        cout.print("[warn]! No logged in user, use [code]pendingai auth login[/].")
        raise Exit(1)
    elif ctx.obj["client"].session.token.is_expired():
        cout.print("[warn]! Access token expired; use [code]pendingai auth refresh[/].")
        raise Exit(1)
    assert ctx.obj["client"].session.info
    email: str = ctx.obj["client"].session.info.user_email
    cout.print(f"[success]✓ Access token for [b underline]{email}[/].")
    cout.print(f"\b{ctx.obj['client'].session.token.access_token}", soft_wrap=True)
